/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
#include "fist.h"
#include "common.h"
#include "tracer.h"
#include "tracefs.h"

// assembly drivers
#include "ad_stream.h"
#include "ad_aggregate.h"

// output drivers
#include "od_file.h"
#include "od_cksum.h"
#include "od_compress.h"
#include "od_encrypt.h"
#include "od_buffer.h"
#include "od_async.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
#define COLOR_WHITE 0
#define COLOR_GREY  1
#define COLOR_BLACK 2


static int do_dfs_validate(struct user_tracecond* root, char *colors, int index, int num_cond)
{
	/* cycle!!! */
	int ft, ff;
	int err = -EINVAL;
	struct user_tracecond *child;
	print_entry_location();
	if (colors[index] == COLOR_GREY) {
		err =  -ELOOP;
		goto out;
	}

	colors[index] = COLOR_GREY;
	ft = root->tc_follow_true;
	ff = root->tc_follow_false;

	if (ft >= num_cond || ff >= num_cond) {
		/* pointers should not go beyond the end of the array */
		goto out;
	} else if (ft < 0 && ff < 0) {
		/* only TRUE or FALSE nodes are allowed to have NULL children */
		if (root->tc_type != TC_TRUE &&  root->tc_type != TC_FALSE) {
			goto out;
		}
	} else if (ft < 0 || ff < 0) {
		/* BOTH pointers should be valid */
		goto out;
	}
	/* pointers are okay, now do DFS on both the left and the right */
	child = root - index + ft;
	if (ft >= 0 && colors[ft] != COLOR_BLACK && (err = do_dfs_validate(root - index + ft, colors, ft, num_cond)) != 0)
		goto out;
	if (ff >= 0 && colors[ff] != COLOR_BLACK && (err = do_dfs_validate(root - index + ff, colors, ff, num_cond)) != 0)
		goto out;
	colors[index] = COLOR_BLACK;
	err = 0;

out:
	print_exit_status(err);
	return err;
}

static void release_trace_conditions(struct super_block *sb, struct trace_destination *dest)
{
    struct list_head *head = &dest->td_conds;
    print_entry_location();
    while (!list_empty(head)) {
	struct tracecond *next = list_entry(head->next, struct tracecond, tc_list);
	list_del(&next->tc_list);
	switch (next->tc_type) {
	    /* do funky stuff per trace condition */
	case TC_NAME:
	case TC_EXT:
	    release_pattern(stopnc(sb), next->u.tc_name);
	    break;

	default: break;



	}
	KFREE(next);
    }
    dest->td_root = NULL;

    print_exit_location();
}


static int process_trace_conditions(struct super_block *sb, struct trace_destination *dest, struct user_tracecond *utcs, int num_conditions)
{
	int err = 0;
	char *colors = NULL;
	struct user_tracecond *tmp;
	struct tracecond *curr = NULL;
	int i;
	struct tracecond **arr = NULL;
	dest->td_root = NULL;
/* do a recursive validation of the input graph */
	print_entry_location();
	if (num_conditions) {
		colors = KMALLOC(num_conditions * sizeof(*colors), GFP_KERNEL);
		if (!colors) {
			err = -ENOMEM;
			goto out;
		}
		memset(colors, 0, num_conditions * sizeof(*colors));
		err = do_dfs_validate(utcs, colors, 0, num_conditions);
	}
	if (err || num_conditions == 0)
	    goto out;

        /* array of tracecondition pointers, for setting up pointers in the graph */
	arr = KMALLOC(num_conditions * sizeof(struct tracecond *), GFP_KERNEL);
	if (!arr) {
	    err = -ENOMEM;
	    goto out;
	}


	ASSERT(list_empty(&dest->td_conds));
	for (i = 0, tmp = utcs; !err && i < num_conditions; i++, tmp++) {
	    curr = KMALLOC(sizeof(struct tracecond), GFP_KERNEL);
	    if (!curr) {
		err = -ENOMEM;
		goto out;
	    }
	    memset(curr, 0, sizeof(*curr));
	    curr->tc_type = tmp->tc_type;
	    switch(tmp->tc_type) {
	    case TC_FALSE:
	    case TC_TRUE:
		break; /* nothing to do */
#ifdef TRACE_SYSCALL
	    case TC_SYSCALL:
#endif
	    case TC_OP:
		memcpy(curr->u.tc_map, tmp->u.tc_map, sizeof(curr->u.tc_map));
		break;
	    case TC_PID:
	    case TC_CUID:
	    case TC_CGID:
	    case TC_FUID:
	    case TC_FGID:
	    case TC_SID:
	    case TC_INO:
	    case TC_PGRP:
		curr->u.tc_ival = tmp->u.tc_ival;
		break;
	    case TC_TIME:
		memcpy(&curr->u.tc_time, &tmp->u.tc_time, sizeof(curr->u.tc_time));
		break;
	    case TC_PNAME:
		memcpy(curr->u.tc_map, tmp->u.tc_map, sizeof(curr->u.tc_map));
		break;
#ifdef TRACE_EXT
	    case TC_EXT:
	    {
                struct qstr str;
                struct name *pat;
                char *name = getname(tmp->u.tc_name);
		if (IS_ERR(name)) {
			err = PTR_ERR(str.name);
			break;
		}
		str.len = strlen(name) + 1;
		str.name = KMALLOC(strlen(name) + 1, GFP_KERNEL);
		if (str.name) {
			char *c = (char *)str.name;
			*c = '.';
			c++;
			memcpy(c, name, str.len - 1);
			if ((pat = add_pattern(stopnc(sb), &str, 1)) != NULL) {
			    curr->u.tc_name = pat;
			} else {
                            err = -ENOMEM;
			}
			KFREE((char *)str.name);
		}
		putname(name);
	    }
            break;
#endif
	    case TC_NAME:
	    {
		struct qstr str;
		struct name* pat;

		str.name = getname(tmp->u.tc_name);
		if (IS_ERR(str.name)) {
		    err = PTR_ERR(str.name);
		    break;
		}
		str.len = strlen(str.name);
		if ((pat = add_pattern(stopnc(sb), &str, 0)) != NULL) {
		    curr->u.tc_name = pat;
		} else {
		    err = -ENOMEM;
		}
		putname(str.name);

	    }
	    break;
	    default:
		fist_dprint(2, "process_trace_conditions: unknown trace condition type = %d\n", tmp->tc_type);
		err = -EINVAL;
	    }
	    if (!err) {
		list_add_tail(&curr->tc_list, &dest->td_conds);
		arr[i] = curr;
	    }
	    curr = NULL;
	}
	if (err)
	    goto out;
	/* now that we've allocated the trace condition nodes, just link the pointers */
	for (i = 0, tmp = utcs; i < num_conditions; i++, tmp++) {

	    if (tmp->tc_follow_true >= 0)
		arr[i]->tc_follow_true = arr[tmp->tc_follow_true];
	    if (tmp->tc_follow_false >= 0)
		arr[i]->tc_follow_false = arr[tmp->tc_follow_false];
	}
	dest->td_root = arr[0];


out:
	if (arr) {
	    KFREE(arr);
	}
	if (curr)
	    KFREE(curr);
	if (err < 0)
	    release_trace_conditions(sb, dest);

	if (colors)
		KFREE(colors);
	print_exit_status(err);
	return err;
}

void destroy_output_driver(struct output_driver *od)
{
    switch (od->od_type)
    {
    case OD_FILE:
	destroy_file_driver(od);
	break;
    case OD_CKSUM:
	destroy_cksum_driver(od);
	break;
    case OD_COMPRESS:
        destroy_compress_driver(od);
	break;
    case OD_ENCRYPT:
	destroy_encrypt_driver(od);
	break;
    case OD_BUFFER:
	destroy_buffer_driver(od);
	break;
    case OD_ASYNC:
	destroy_async_driver(od);
	break;
    default:
	ASSERT(0);
    }
}

static void release_output_drivers(struct trace_destination *dest)
{
    struct output_driver *tmp;

    print_entry_location();

    while (dest->td_od) {
	dest->td_od->od_ops->release(dest->td_od);
	tmp = dest->td_od->od_next;
	destroy_output_driver(dest->td_od);
	dest->td_od = tmp;
    }

    print_exit_location();
}



static void destroy_assembly_driver(struct assembly_driver *ad)
{
	switch (ad->ad_type)
	{
		case AD_STREAM:
			destroy_stream_driver(ad);
			break;
		case AD_AGGREGATE:
			destroy_aggregate_driver(ad);
			break;
		default:
			ASSERT(0);
	}
}


static void release_assembly_drivers(struct trace_destination *dest)
{
	struct assembly_driver *tmp;
	print_entry_location();

	while (dest->td_ad) {
	    dest->td_ad->ad_ops->release(dest->td_ad);
	    tmp = dest->td_ad->ad_next;
	    destroy_assembly_driver(dest->td_ad);
	    dest->td_ad = tmp;
	}

	print_exit_location();
}





int tracefs_ioctl_add_odrivers(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg) {
	unsigned int n_drivers;
	int err = 0;
	struct user_output_driver *ods = NULL;
	struct output_driver *tmp;
	int i, j;
	int destid = -1;

	print_entry_location();

	if (get_user(n_drivers, (unsigned int *)arg)) {
		err = -EFAULT;
		goto out;
	}

	arg = (unsigned long)(((unsigned char *)arg) + (sizeof (unsigned int)));
	ods = KMALLOC(n_drivers * sizeof(struct user_output_driver), GFP_KERNEL);
	if (!ods) {
		err = -ENOMEM;
		goto out;
	}

	if (copy_from_user(ods, (unsigned char *) arg, n_drivers * sizeof(struct user_output_driver))) {
		err = -EFAULT;
		goto out;
	}

	lock_super(inode->i_sb);
	for (j = 0; j < MAX_DESTINATIONS && destid == -1; j++) {
		if(stoptr(inode->i_sb)->dest[j].td_state == TD_FREE) {
			destid = j;
		}
	}
	if (destid == -1) {
		err = -ENOBUFS;
		goto out;
	}
	stoptr(inode->i_sb)->dest[destid].td_state = TD_CONSTRUCTING;
	unlock_super(inode->i_sb);
	INIT_LIST_HEAD(&stoptr(inode->i_sb)->dest[destid].td_conds);

	/* create output drivers one by one. Only the output drivers
 	 * that get totally initialised make it into the output driver
 	 * chain.
         */
	for (i = 0; i < n_drivers; i++) {
		j = n_drivers - i - 1;
		switch(ods[i].od_type) {
		case OD_ENCRYPT:
			tmp = create_encrypt_driver();
			if (!tmp) {
				err = -ENOMEM;
				goto out;
			}
			if ((err = tmp->od_ops->init(tmp, &ods[i].u.od_encrypt, stoptr(inode->i_sb)->dest[destid].td_od, j)) != 0) {
				goto out;
			}
			break;
		case OD_CKSUM:
			tmp = create_cksum_driver();
			if (!tmp) {
				err = -ENOMEM;
				goto out;
			}
			if ((err = tmp->od_ops->init(tmp, &ods[i].u.od_cksum, stoptr(inode->i_sb)->dest[destid].td_od, j)) != 0) {
				goto out;
			}
			break;
		case OD_COMPRESS:
			tmp = create_compress_driver();
			if (!tmp) {
				err = -ENOMEM;
				goto out;
			}
			if ((err = tmp->od_ops->init(tmp, &ods[i].u.od_compress, stoptr(inode->i_sb)->dest[destid].td_od, j)) != 0) {
				goto out;
			}
			break;
		case OD_FILE:
			tmp = create_file_driver();
			if (!tmp) {
				err = -ENOMEM;
				goto out;
			}
			if ((err = tmp->od_ops->init(tmp, &ods[i].u.od_file, stoptr(inode->i_sb)->dest[destid].td_od, j)) != 0) {
			    goto out;
			}

			break;
		case OD_BUFFER:
			tmp = create_buffer_driver();
			if (!tmp) {
				err = -ENOMEM;
				goto out;
			}
			if ((err = tmp->od_ops->init(tmp, &ods[i].u.od_buffer, stoptr(inode->i_sb)->dest[destid].td_od, j)) != 0) {
			    goto out;
			}
			break;
		case OD_ASYNC:
			tmp = create_async_driver();
			if (!tmp) {
				err = -ENOMEM;
				goto out;
			}
			if ((err = tmp->od_ops->init(tmp, &ods[i].u.od_async, stoptr(inode->i_sb)->dest[destid].td_od, j)) != 0) {
			    goto out;
			}

			break;
		default:
			err = -EINVAL;
			goto out;
		}
		stoptr(inode->i_sb)->dest[destid].td_od = tmp;
	}
	/* all good, now change state */
	lock_super(inode->i_sb);
	mntget(file->f_vfsmnt);
	stoptr(inode->i_sb)->dest[destid].td_state = TD_ODRIVER;
	unlock_super(inode->i_sb);

out:
	if (err < 0) {
		if (destid != -1) {
			/* Do the releasing of Output drivers that have been setup. */
			release_output_drivers(&(stoptr(inode->i_sb)->dest[destid]));

			lock_super(inode->i_sb);
			stoptr(inode->i_sb)->dest[destid].td_state = TD_FREE;
			unlock_super(inode->i_sb);
		}
	}
	if (ods) {
		KFREE(ods);
	}

	err = (err == 0) ? destid : err;
	print_exit_status(err);
	return err;
}

int tracefs_ioctl_add_adrivers(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
	unsigned int n_drivers;
	int err = 0;
	struct user_assembly_driver *ads = NULL;
	struct assembly_driver *tmp;
	struct trace_sb *td = stoptr(inode->i_sb);
	int i;
	int dest_id = -1;

	print_entry_location();

	/* get the destination id */
	if (get_user(dest_id, (unsigned int *)arg)) {
		err = -EFAULT;
		goto out_retain_state;
	}

	/* validate the destination id first */
	/* Also check that the output drivers have been added previously */
	if (dest_id < 0 || dest_id >= MAX_DESTINATIONS) {
		err = -EINVAL;
		goto out_retain_state;
	}

	lock_super(inode->i_sb);
	if (td->dest[dest_id].td_state != TD_ODRIVER) {
		err = -EINVAL;
		unlock_super(inode->i_sb);
		goto out_retain_state;
	}
	td->dest[dest_id].td_state = TD_CONSTRUCTING;
	unlock_super(inode->i_sb);


	arg = (unsigned long)(((unsigned char *)arg) + (sizeof (unsigned int)));


	if (get_user(n_drivers, (unsigned int *)arg)) {
		err = -EFAULT;
		goto out;
	}


	arg = (unsigned long)(((unsigned char *)arg) + (sizeof (unsigned int)));
	ads = KMALLOC(n_drivers * sizeof(struct user_assembly_driver), GFP_KERNEL);

	if (!ads) {
		err = -ENOMEM;
		goto out;
	}

	if (copy_from_user(ads, (unsigned char *)arg, n_drivers * sizeof(struct user_assembly_driver))) {
		err = -EFAULT;
		goto out;
	}

	for (i = 0; i < n_drivers; i++) {

		switch(ads[i].ad_type) {
			case AD_STREAM:
				tmp = create_stream_driver();
				if (!tmp) {
					err = -ENOMEM;
					goto out;
				}
				if ((err = tmp->ad_ops->init(tmp, (void *)ads[i].u.ad_stream.ad_verbosity)) != 0) {
					goto out;
				}

				td->dest[dest_id].td_ad = tmp;
				break;
			case AD_AGGREGATE:
				tmp = create_aggregate_driver();
				if (!tmp) {
					err = -ENOMEM;
					goto out;
				}
				if ((err = tmp->ad_ops->init(tmp, NULL)) != 0) {
					goto out;
				}
				td->dest[dest_id].td_ad = tmp;
				break;
			default:
				err = -EINVAL;
				goto out;
		}
	}

	/* all good, change state */
	lock_super(inode->i_sb);
	td->dest[dest_id].td_state = TD_ADRIVER;
	unlock_super(inode->i_sb);


out:
	if (err < 0) {
		if (dest_id != -1) {
			release_assembly_drivers(&td->dest[dest_id]);
			lock_super(inode->i_sb);
			td->dest[dest_id].td_state = TD_ODRIVER;
			unlock_super(inode->i_sb);
		}
	}
out_retain_state:
	if (ads)
		KFREE(ads);
	print_exit_status(err);
	return err;
}


int tracefs_ioctl_add_rule(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg) {

    int err = 0;
    unsigned int n_rules;
    struct user_tracecond *utc = NULL;
    struct trace_sb *td = stoptr(inode->i_sb);
    int dest_id = -1;

    print_entry_location();

    /* get the destination id */
    if (get_user(dest_id, (unsigned int *)arg)) {
	err = -EFAULT;
	goto out_retain_state;
    }

/* validate the destination id first */

    if (dest_id < 0 || dest_id >= MAX_DESTINATIONS) {
	err = -EINVAL;
	goto out_retain_state;
    }
/* Also check that the output drivers and assembly drivers have been added previously */

    lock_super(inode->i_sb);
    if (td->dest[dest_id].td_state != TD_ADRIVER) {
	err = -EINVAL;
	unlock_super(inode->i_sb);
	goto out_retain_state;
    }
    td->dest[dest_id].td_state = TD_CONSTRUCTING;
    unlock_super(inode->i_sb);

    arg = (unsigned long)(((unsigned char *)arg) + (sizeof (unsigned int)));


    if (get_user(n_rules, (unsigned int *)arg)) {
	err = -EFAULT;
	goto out;
    }

    arg = (unsigned long)(((unsigned char *)arg) + (sizeof (unsigned int)));
    utc = KMALLOC(n_rules * sizeof(struct user_tracecond), GFP_KERNEL);

    if (!utc) {
	err = -ENOMEM;
	goto out;
    }

    if (copy_from_user(utc, (unsigned char *)arg, n_rules * sizeof(struct user_tracecond))) {
	err = -EFAULT;
	goto out;
    }


    err = process_trace_conditions(inode->i_sb, &td->dest[dest_id], utc, n_rules);
    if (!err) {
	lock_super(inode->i_sb);
	td->dest[dest_id].td_state = TD_DISABLED;
	unlock_super(inode->i_sb);
    }
  out:
    if (err < 0) {
	lock_super(inode->i_sb);
	td->dest[dest_id].td_state = TD_ADRIVER;
	unlock_super(inode->i_sb);
    }

  out_retain_state:

    if (utc)
	KFREE(utc);
    print_exit_status(err);
    return err;
}

int tracefs_ioctl_del_odrivers(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg) {
	int err = 0;
		int dest_id = (int)arg;
	struct trace_sb *td = stoptr(inode->i_sb);
	print_entry_location();
	if (dest_id < 0 || dest_id >= MAX_DESTINATIONS) {
	    err = -EINVAL;
	    goto out;
	}

	lock_super(inode->i_sb);
	if (td->dest[dest_id].td_state != TD_ODRIVER) {
	    err = -EINVAL;
	    unlock_super(inode->i_sb);
	    goto out;
	}
	td->dest[dest_id].td_state = TD_CONSTRUCTING;
	unlock_super(inode->i_sb);
	release_output_drivers(&td->dest[dest_id]);

	lock_super(inode->i_sb);
	mntput(file->f_vfsmnt);
	td->dest[dest_id].td_state = TD_FREE;
	unlock_super(inode->i_sb);

out:
	print_exit_status(err);
	return err;
}

int tracefs_ioctl_del_adrivers(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
	int err = 0;
	int dest_id = (int)arg;
	struct trace_sb *td = stoptr(inode->i_sb);
	print_entry_location();

	if (dest_id < 0 || dest_id >= MAX_DESTINATIONS) {
	    err = -EINVAL;
	    goto out;
	}

	lock_super(inode->i_sb);
	if (td->dest[dest_id].td_state != TD_ADRIVER) {
	    err = -EINVAL;
	    unlock_super(inode->i_sb);
	    goto out;
	}
	td->dest[dest_id].td_state = TD_CONSTRUCTING;
	unlock_super(inode->i_sb);
	release_assembly_drivers(&td->dest[dest_id]);

	lock_super(inode->i_sb);
	td->dest[dest_id].td_state = TD_ODRIVER;
	unlock_super(inode->i_sb);

out:
	print_exit_status(err);
	return err;
}

int tracefs_ioctl_del_rule(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
	int err = 0;
	int dest_id = (int)arg;
	struct trace_sb *td = stoptr(inode->i_sb);
	print_entry_location();
	if (dest_id < 0 || dest_id >= MAX_DESTINATIONS) {
	    err = -EINVAL;
	    goto out;
	}

	lock_super(inode->i_sb);
	if (td->dest[dest_id].td_state != TD_DISABLED && td->dest[dest_id].td_state != TD_ERROR) {
	    err = -EINVAL;
	    unlock_super(inode->i_sb);
	    goto out;
	}
	td->dest[dest_id].td_state = TD_CONSTRUCTING;
	unlock_super(inode->i_sb);
	release_trace_conditions(inode->i_sb, &td->dest[dest_id]);

	lock_super(inode->i_sb);
	td->dest[dest_id].td_state = TD_ADRIVER;
	unlock_super(inode->i_sb);

out:
	print_exit_status(err);
	return err;
}

int tracefs_ioctl_teardown(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
	int err = 0;
	int dest_id = (int)arg;
	struct trace_sb *td = stoptr(inode->i_sb);
	struct trace_destination *dest;
	int old_state;
	print_entry_location();

	if (dest_id < 0 || dest_id >= MAX_DESTINATIONS) {
	    err = -EINVAL;
	    goto out;
	}
	dest = &td->dest[dest_id];

	lock_super(inode->i_sb);
	if (dest->td_state != TD_DISABLED
	    && dest->td_state != TD_ERROR
	    && dest->td_state != TD_ADRIVER
	    && dest->td_state != TD_ODRIVER) {
	    err = -EINVAL;
	    unlock_super(inode->i_sb);
	    goto out;
	}
	old_state = dest->td_state;
	dest->td_state = TD_CONSTRUCTING;
	unlock_super(inode->i_sb);

	switch (old_state) {
	case TD_DISABLED:
	case TD_ERROR:
	    release_trace_conditions(inode->i_sb, dest);
	    /* FALL THROUGH!!! */
	case TD_ADRIVER:
	    release_assembly_drivers(dest);
	    /* FALL THROUGH!!! */
	case TD_ODRIVER:
	    release_output_drivers(dest);
	}

	lock_super(inode->i_sb);
	mntput(file->f_vfsmnt);
	td->dest[dest_id].td_state = TD_FREE;
	unlock_super(inode->i_sb);

out:
	print_exit_status(err);
	return err;
}

int tracefs_ioctl_start_trace(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
    	int err = 0;
	int dest_id;
	struct trace_sb *td = stoptr(inode->i_sb);
	print_entry_location();
	dest_id = (int)arg;

	if (dest_id < 0 || dest_id >= MAX_DESTINATIONS) {
	    err = -EINVAL;
	    goto out;
	}

	lock_super(inode->i_sb);
	if (td->dest[dest_id].td_state != TD_DISABLED) {
	    err = -EINVAL;
	    unlock_super(inode->i_sb);
	    goto out;
	}

	td->dest[dest_id].td_state = TD_ENABLED;
	unlock_super(inode->i_sb);
	/* write the start of trace message */
	td->dest[dest_id].td_ad->ad_ops->ao_start_trace(td->dest[dest_id].td_ad, &td->dest[dest_id]);



out:
	print_exit_status(err);
	return err;

}

int tracefs_ioctl_stop_trace(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
    	int err = 0;
	int dest_id = (int)arg;
	struct trace_sb *td = stoptr(inode->i_sb);
	print_entry_location();

	if (dest_id < 0 || dest_id >= MAX_DESTINATIONS) {
	    err = -EINVAL;
	    goto out;
	}

	lock_super(inode->i_sb);
	if (td->dest[dest_id].td_state != TD_ENABLED) {
	    err = -EINVAL;
	    unlock_super(inode->i_sb);
	    goto out;
	}
	td->dest[dest_id].td_state = TD_CONSTRUCTING; // transition state
	unlock_super(inode->i_sb);

	/* write the stop of trace message */
	td->dest[dest_id].td_ad->ad_ops->ao_stop_trace(td->dest[dest_id].td_ad, &td->dest[dest_id]);
	/* flush */
	td->dest[dest_id].td_od->od_ops->flush(td->dest[dest_id].td_od);

	lock_super(inode->i_sb);
	td->dest[dest_id].td_state = TD_DISABLED;
	unlock_super(inode->i_sb);


out:
	print_exit_status(err);
	return err;

}
