/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * od_cksum.c
 * output driver for checksum
 */
#include "tracer.h"
#include "od_cksum.h"
#include "md5.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
#define MD5_CKSUM_SIZE (16 + sizeof(int))
/*
 * internal structure for this driver
 */

struct od_cksum_data {

	char *buf;     /* buffer */
	char *ptr; /* remaining bytes in buffer */
	char *begin; /* start of the "real" buffer", this remains constant */
	int size; /* size of the buffer */
	block_index_t block_index; /* index of the block */
	md5_state_t state; /* md5 state */
	unsigned char k_ipad[65];
	unsigned char k_opad[65];

};

/**
 * Destroys the checksum driver object.
 * Make sure that release function has already been called
 * @param output driver to be destroyed
 */

void destroy_cksum_driver(struct output_driver *od)
{
	ASSERT(od);
	ASSERT(od->od_private_data == NULL);
	/* nothing to do, as of now */
	KFREE(od);
}

/**
 * Initializes the file driver object.
 * @param Checksum driver object to initialize
 * @param <code>struct od_cksum_params</code> type initialization parameters
 * @see struct od_cksum_params
 */

static int cksum_init(struct output_driver *this, void *params, struct output_driver *next, int level)
{
	int ret = 0;
	struct od_cksum_data *cdata;
	char *ptr;
	char *key = NULL;
	int i;
	struct od_cksum_params *cp = (struct od_cksum_params *)params;

	ASSERT(this->od_private_data == NULL);
	print_entry_location();
	if (!cp || !next || cp->keylen > 64) {
		ret = -EINVAL;
		goto out;
	}

	cdata = KMALLOC(sizeof(struct od_cksum_data), GFP_KERNEL);
	if (!cdata) {
		ret = -ENOMEM;
		goto out;
	}

	cdata->buf = (char *)__get_free_page(GFP_KERNEL);
	if (!cdata->buf) {
		KFREE(cdata);
		ret = -ENOMEM;
		goto out;
	}

	if ((key = KMALLOC(cp->keylen, GFP_KERNEL)) == NULL) {
		ret = -ENOMEM;
		goto out;
	}

	if (copy_from_user(key, cp->key, cp->keylen)) {
		ret = -EFAULT;
		goto out;
	}

	cdata->begin = cdata->ptr = cdata->buf + MD5_CKSUM_SIZE;
	cdata->size = PAGE_SIZE - MD5_CKSUM_SIZE;
	cdata->block_index = 0;
	ptr = cdata->begin - sizeof(cdata->block_index);

	memset(cdata->k_ipad, 0, sizeof(cdata->k_ipad));
	memset(cdata->k_opad, 0, sizeof(cdata->k_opad));

	memcpy(cdata->k_ipad, key, cp->keylen);
	memcpy(cdata->k_opad, key, cp->keylen);

	for (i = 0; i < 64; i++) {
		cdata->k_ipad[i] ^= 0x36;
		cdata->k_opad[i] ^= 0x5c;
	}

	md5_init(&cdata->state);
	md5_append(&cdata->state, cdata->k_ipad, 64);
	memcpy(ptr, &cdata->block_index, sizeof(cdata->block_index));
	md5_append(&cdata->state, ptr, sizeof(cdata->block_index));
	cdata->block_index++;
	this->od_private_data = cdata;

	this->od_next = next;

out:
	if (key)
		KFREE(key);
	print_exit_status(ret);
	return ret;
}

/**
 * Releases the private data.
 * Flush should be called before this. Also causes the chained drivers
 * to be released.
 * @param driver to release
 */
void cksum_release(struct output_driver *this)
{
	struct od_cksum_data *cdata = (struct od_cksum_data *)this->od_private_data;
	ASSERT(this);
	ASSERT(this->od_type == OD_CKSUM);
	free_page((unsigned long)cdata->buf);
	KFREE(cdata);
	this->od_private_data = NULL;
}

/**
 * flushes the internal buffer, if any.
 * @param object to operate on
 * @return Number of bytes written.
 */
int cksum_flush(struct output_driver *this)
{
	struct od_cksum_data *cdata = (struct od_cksum_data *)this->od_private_data;
	int ret = 0;

	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_CKSUM);
	ASSERT(this->od_next);

	if (cdata->ptr - cdata->begin != 0) {
		md5_finish(&cdata->state, cdata->buf);
		md5_init(&cdata->state);
		md5_append(&cdata->state, cdata->k_opad, 64);
		md5_append(&cdata->state, cdata->buf, 16);
		md5_finish(&cdata->state, cdata->buf);
		ret = this->od_next->od_ops->write(this->od_next, cdata->buf, cdata->ptr - cdata->buf,OO_KERNEL_BUFFER);
	}
	if (ret >= 0) {
		ret = this->od_next->od_ops->flush(this->od_next);
	}

	memset(cdata->k_ipad, 0, sizeof(cdata->k_ipad));
	memset(cdata->k_opad, 0, sizeof(cdata->k_opad));

	print_exit_status(ret);
	return ret;
}

/**
 * Writes the data to output.
 * If internal buffering is enabled, data is written to buffer
 * till the buffer runs out.
 * @param object to operate on
 * @param buffer to write out
 * @param number of bytes to write
 * @param whether the buffer is a user-level buffer (OO_USER_BUFFER)
 * @returns number of bytes written
 */

int cksum_write(struct output_driver *this, void *buf, int len,int is_user_buffer)
{
	struct od_cksum_data *cdata = (struct od_cksum_data *)this->od_private_data;
	int ret = 0;
	int remaining;
	char *tmp;
	int total;
	int junk;
	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_CKSUM);
	/* Checksum driver cannot work in isolation */
	ASSERT(this->od_next);
	tmp = (char *)buf;
	total = len;
	remaining = cdata->begin + cdata->size - cdata->ptr;
	while(remaining < len) {
         	if(is_user_buffer == OO_KERNEL_BUFFER)
		  memcpy(cdata->ptr, tmp, remaining);
		else
		  junk = copy_from_user(cdata->ptr, tmp, remaining);
		len -= remaining;
		/* cksum the incoming data */
		md5_append(&cdata->state, cdata->ptr, remaining);
		md5_finish(&cdata->state, cdata->buf);
		md5_init(&cdata->state);
		md5_append(&cdata->state, cdata->k_opad, 64);
		md5_append(&cdata->state, cdata->buf, 16);
		md5_finish(&cdata->state, cdata->buf);

		tmp += remaining;
		ret = this->od_next->od_ops->write(this->od_next, cdata->buf, cdata->size + MD5_CKSUM_SIZE,OO_KERNEL_BUFFER);
		if (ret < 0)
			goto out;
		memcpy(cdata->begin - sizeof(cdata->block_index), &cdata->block_index, sizeof(cdata->block_index));
		md5_init(&cdata->state);
		md5_append(&cdata->state, cdata->k_ipad, 64);
		md5_append(&cdata->state, (void *)&cdata->block_index, sizeof(cdata->block_index));
		cdata->block_index++;
		remaining = cdata->size;
		cdata->ptr = cdata->begin;

	}
	if (len > 0) {
	        if(is_user_buffer == OO_KERNEL_BUFFER)
		  memcpy(cdata->ptr, tmp, len);
		else
		  junk = copy_from_user(cdata->ptr, tmp, len);
		cdata->ptr += len;
		md5_append(&cdata->state, cdata->ptr, len);
	}
	ret = total;

out:
	print_exit_status(ret);
	return ret;
}

/**
 * Returns the preferred block size
 * @param object to operate on
 * @returns block size
 */
int cksum_getblocksize(struct output_driver *this)
{
	ASSERT(this);
	ASSERT(this->od_type == OD_CKSUM);
	return PAGE_SIZE;
}

static struct output_ops cksum_od_ops = {
write: cksum_write,
flush: cksum_flush,
getblocksize: cksum_getblocksize,
init: cksum_init,
release: cksum_release
};

/**
 * Creates a file driver object.
 * @return A file output driver object with function pointers initialised
 */

struct output_driver* create_cksum_driver()
{
	struct output_driver *driver = (struct output_driver *)KMALLOC(sizeof(struct output_driver), GFP_KERNEL);

	if (driver) {
		driver->od_ops = &cksum_od_ops;
		driver->od_type = OD_CKSUM;
		driver->od_private_data = NULL;
		driver->od_next = NULL;
	}
	return driver;
}


