/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * od_file.c
 * output driver for file based output
 */

#include "tracer.h"
#include "od_file.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
/*
 * internal structure for this driver
 */

struct od_file_data {
	struct file *file;
	char *buf;     /* buffer */
	int remaining; /* remaining bytes in buffer */
	int buf_size;
};

/**
 * Destroys the file driver object.
 * Make sure that release function has already been called
 * @param output driver to be destroyed
 */

void destroy_file_driver(struct output_driver *od)
{
	ASSERT(od);
	ASSERT(od->od_private_data == NULL);
	/* nothing to do, as of now */
	KFREE(od);
}

/**
 * Initializes the file driver object.
 * @param File driver object to initialize
 * @param <code>struct od_file_params</code> type initialization parameters
 * @see struct od_file_params
 */

static int file_init(struct output_driver *this, void *params, struct output_driver *next, int level)
{
	int ret = -ENOMEM;
	struct od_file_data *fdata;
	struct file *file = NULL;
	char *name;
	struct od_file_params *fp = (struct od_file_params *)params;
	ASSERT(this->od_private_data == NULL);

	print_entry_location();

	if (!fp || next) {
		ret = -EINVAL;
		goto out;
	}

	if (fp->name == NULL) {
		/* File descriptor mode of operation
		 * fget the fd from the current process's
		 * file descriptor table
		 */
		if ((file = fget(fp->fd)) == NULL) {
			ret = -EINVAL;
			goto out;
		}
		if (atomic_read(&file->f_count) != 2) {
			ret = -EBUSY;
			goto out_close;
		}
	} else {

		name = getname(fp->name);
		if (IS_ERR(name)) {
			ret = PTR_ERR(name);
			goto out;
		}
		//file = filp_open(name, O_CREAT | O_WRONLY | O_EXCL, 0644);
		if (fp->sync_writes)
			file = filp_open(name, O_CREAT | O_WRONLY | O_TRUNC | O_SYNC | O_LARGEFILE, 0644);
		else
			file = filp_open(name, O_CREAT | O_WRONLY | O_TRUNC | O_LARGEFILE, 0644);
		putname(name);
		if (IS_ERR(file)) {
			ret = PTR_ERR(file);
			goto out;
		}
		if (atomic_read(&file->f_count) != 1) {
			ret = -EBUSY;
			goto out_close;
		}
	}

	fdata = this->od_private_data = KMALLOC(sizeof(struct od_file_data), GFP_KERNEL);
	if (!fdata)
		goto out_close;
	memset(this->od_private_data, 0, sizeof(struct od_file_data));
	/* allocate the buffer */
	fdata->remaining = fdata->buf_size = fp->buf_size;
	fdata->file = file;
	if (fp->buf_size) {
		printk("asking for %d\n", fp->buf_size);
		fdata->buf = (char *)vmalloc(fp->buf_size);
		if (!fdata->buf)
			goto out_free;
	}
	/* all good! */
	ret = 0;
	goto out;

out_free:
	KFREE(fdata);
	this->od_private_data = NULL;

out_close:
	if (file)
		fput(file);
out:
	print_exit_status(ret);
	return ret;
}

/**
 * Releases the private data.
 * Flush should be called before this. Also causes the chained drivers
 * to be released.
 * @param driver to release
 */
void file_release(struct output_driver *this)
{
	struct od_file_data *fdata;
	ASSERT(this);
	ASSERT(this->od_type == OD_FILE);
	fdata = this->od_private_data;
	/* call release on chained driver */
	//if (this->next)
	//	this->next->od_ops->release(this->next);
	/* close the file */
	fput(fdata->file);
	/* free the buffer, if any */
	vfree(fdata->buf);
	KFREE(fdata);
	this->od_private_data = NULL;
}

static inline int do_file_write(struct file *file, void *buf, int len)
{
	int ret = 0;
	mm_segment_t oldfs;
	print_entry_location();
	oldfs = set_fs(KERNEL_DS);

	/*
	 * Tim: this seems to prevent a deadlock in the logfile filesystem
	 * while logging writepage operations.  It may be inappropriate
	 * to call a f_op->write() during writepage; especially under
	 * heavy memory pressure.
	 */
	cond_resched();

	ret = file->f_op->write(file, buf, len, &file->f_pos);
	set_fs(oldfs);
	print_exit_status(ret);
	return ret;
}

/**
 * flushes the interbal buffer, if any.
 * @param object to operate on
 * @return Number of bytes written.
 */
int file_flush(struct output_driver *this)
{
	struct od_file_data *fdata;
	int ret = 0;

	print_entry_location();
	ASSERT(this);
	ASSERT(this->od_type == OD_FILE);
	fdata = this->od_private_data;
	if (fdata->remaining < fdata->buf_size) {
	    /* This is hypothetical. There usually will not be a chained output driver for this driver */
	    if (this->od_next)
		ret = this->od_next->od_ops->write(this->od_next, fdata->buf, fdata->buf_size - fdata->remaining,OO_KERNEL_BUFFER);
	    else
		ret = do_file_write(fdata->file, fdata->buf, fdata->buf_size - fdata->remaining);
	    fdata->remaining = fdata->buf_size;
	}
	print_exit_status(ret);
	return ret;
}

/**
 * Writes the data to output.
 * If internal buffering is enabled, data is written to buffer
 * till the buffer runs out.
 * @param object to operate on
 * @param buffer to write out
 * @param number of bytes to write
 * @param whether the buffer is a user-level buffer (OO_USER_BUFFER)
 * @returns number of bytes written
 */

int file_write(struct output_driver *this, void *buf, int len,int is_user_buffer)
{
    int ret = 0;
    int total;
    int junk;
    struct od_file_data *fdata;
    print_entry_location();
    ASSERT(this);
    ASSERT(this->od_type == OD_FILE);
    fdata = this->od_private_data;

    if (fdata->buf_size) { /* buffering is enabled */
	char *tmp = (char *)buf;
	total = len;
	while(fdata->remaining < len) {
	    len -= fdata->remaining;
	    if(is_user_buffer == OO_KERNEL_BUFFER)
	      memcpy(fdata->buf + fdata->buf_size - fdata->remaining, tmp, fdata->remaining);
	    else
	      junk = copy_from_user(fdata->buf + fdata->buf_size - fdata->remaining, tmp, fdata->remaining);
	    tmp += fdata->remaining;
	    if (this->od_next)
		ret = this->od_next->od_ops->write(this->od_next, fdata->buf, fdata->buf_size,OO_KERNEL_BUFFER);
	    else
		ret = do_file_write(fdata->file, fdata->buf, fdata->buf_size);
	    if (ret < 0)
		goto out;
	    fdata->remaining = fdata->buf_size;
	}
	if (len > 0) {
	    if(is_user_buffer == OO_KERNEL_BUFFER)
	      memcpy(fdata->buf + fdata->buf_size - fdata->remaining, tmp, len);
	    else
	      junk = copy_from_user(fdata->buf + fdata->buf_size - fdata->remaining, tmp, len);
	    fdata->remaining -= len;
	}
	ret = total;

    } else {
	if (this->od_next)  /* hypothetical */
	    ret = this->od_next->od_ops->write(this->od_next, buf, len,OO_KERNEL_BUFFER);
	else
	    ret = do_file_write(fdata->file, buf, len);
    }

  out:
    print_exit_status(ret);
    return ret;
}

/**
 * Returns the preferred block size
 * @param object to operate on
 * @returns block size
 */
int file_getblocksize(struct output_driver *this)
{
    struct od_file_data *fdata;
    ASSERT(this);
    ASSERT(this->od_type == OD_FILE);
    fdata = this->od_private_data;
    if (fdata->buf_size)
	return fdata->buf_size;
    else
	return PAGE_SIZE;
}

static struct output_ops file_od_ops = {
    write: file_write,
    flush: file_flush,
    getblocksize: file_getblocksize,
    init: file_init,
    release: file_release
};

/**
 * Creates a file driver object.
 * @return A file output driver object with function pointers initialised
 */

struct output_driver* create_file_driver(void)
{
    struct output_driver *driver = (struct output_driver *)KMALLOC(sizeof(struct output_driver), GFP_KERNEL);

    if (driver) {
	driver->od_ops = &file_od_ops;
	driver->od_type = OD_FILE;
	driver->od_private_data = NULL;
	driver->od_next = NULL;
    }
    return driver;
}


